/*
 * Copyright 2007, Thomas Maier-Komor
 * Licenses: LGPL, see file lgpl.txt
 */

#ifndef _REENTRANT
#define _REENTRANT
#endif
#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <link.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/varargs.h>


typedef enum { cl/*close*/, wr/*write*/, tr/*terminate*/ } task_t;

typedef struct op
{
	task_t op;
	char *buf;
	size_t s;
	struct op *next;
} op_t;

typedef struct operator
{
	int fd;
	char *fn;
	pthread_t thread;
	op_t *volatile head, *volatile tail;
	pthread_mutex_t lock;
	sem_t qlen;
	struct operator *next;
} operator_t;

static operator_t *Operators = 0;
static void *libc_hdl = 0;
static int Terminate = 0, Verbose = 0;
static int MaxThreads = 16, NumThreads = 0;
static sem_t Idle;
static int (*realclose)(int) = 0,
	   (*realwrite)(int, const char *, size_t) = 0,
	   (*realread)(int, char *, size_t) = 0,
	   (*realopen)(const char *, unsigned, unsigned) = 0,
	   (*realopen64)(const char *, unsigned, unsigned) = 0,
	   (*realopenat)(int, const char *, unsigned, unsigned) = 0,
	   (*realopenat64)(int, const char *, unsigned, unsigned) = 0;
static void (*realexit)(int) = 0;
static void *writer_thread(void *);
static void debug(const char *str, ...);
static void log(const char *str, ...);



static void init(void)
{
	char *V,*M;
	int err;
	if (libc_hdl != 0)
		return;
	libc_hdl = dlopen("libc.so",RTLD_NOW|RTLD_LOCAL);
	assert(libc_hdl);
	realopen64 = (int(*)(const char *,unsigned,unsigned)) dlsym(libc_hdl,"open64");
	assert(realopen64);
	realopen = (int(*)(const char *,unsigned,unsigned)) dlsym(libc_hdl,"open");
	assert(realopen);
	realopenat64 = (int(*)(int,const char *,unsigned,unsigned)) dlsym(libc_hdl,"openat64");
	assert(realopenat64);
	realopenat = (int(*)(int,const char *,unsigned,unsigned)) dlsym(libc_hdl,"openat");
	assert(realopenat);
	realclose = (int(*)(int)) dlsym(libc_hdl,"close");
	assert(realclose);
	realread = (int(*)(int, char *, size_t)) dlsym(libc_hdl,"read");
	assert(realread);
	realwrite = (int(*)(int, const char *, size_t)) dlsym(libc_hdl,"write");
	assert(realwrite);
	realexit = (void(*)(int)) dlsym(libc_hdl,"exit");
	assert(realexit);
	err = sem_init(&Idle,0,0);
	assert(err != -1);
	if (V = getenv("MTWRITE_VERBOSE"))
		Verbose = atol(V);
	if (M = getenv("MTWRITE_MAXTHREADS"))
		MaxThreads = atol(M);
	if (Verbose)
		log(	"\nMT-write version r70125\n"
			"Copyright 2007, Thomas Maier=Komor\n"
			"Licenses: LGPL, see file lgpl.txt for details\n"
			"=============================================\n");
	debug("init successful(Verbose = %d,MaxThreads = %d)\n",Verbose,MaxThreads);
}


static void error(const char *str, ...)
{
	char msg[256] = "mtwrite.so: ";
	va_list val;
	va_start(val,str);
	assert(libc_hdl != 0);
	(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
	(void) realwrite(2/*STDERR_FILENO*/,msg,strlen(msg));
	va_end(val);
}


static void log(const char *str, ...)
{
	if (Verbose >= 1) {
		char msg[256] = "mtwrite.so: ";
		va_list val;
		va_start(val,str);
		assert(libc_hdl != 0);
		(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
		(void) realwrite(2/*STDERR_FILENO*/,msg,strlen(msg));
		va_end(val);
	}
}


static void debug(const char *str, ...)
{
	if (Verbose >= 2) {
		va_list val;
		va_start(val,str);
		char msg[256] = "mtwrite.so: ";
		init();
		(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
		(void) realwrite(2/*STDERR_FILENO*/,msg,strlen(msg));
		va_end(val);
	}
}


static void assignthread(int ret,const char *fn)
{
	int err;
	operator_t *op = Operators;
	if (op) {
		if (NumThreads == MaxThreads) {
			err = sem_wait(&Idle);
			assert(err == 0);
		}
		do {
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			assert(op->fd != ret);
			if ((op->head == 0) && (op->fd == -1)) {
				/* thread went out of work */
				op->fd = ret;
				op->fn = strdup(fn);
				err = pthread_mutex_unlock(&op->lock);
				assert(err == 0);
				if (NumThreads != MaxThreads) {
					err = sem_wait(&Idle);
					assert(err == 0);
				}
				debug("open: reusing thread %d\n",op->thread);
				return;
			}
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			op = op->next;
		} while (op);
	}
	op = malloc(sizeof(operator_t));
	op->fd = ret;
	op->fn = strdup(fn);
	err = pthread_mutex_init(&op->lock,0);
	assert(err == 0);
	err = sem_init(&op->qlen,0,0);
	assert(err == 0);
	op->head = op->tail = 0;
	err = pthread_create(&op->thread,0,writer_thread,op);
	assert(err == 0);
	++NumThreads;
	debug("open: created thread %d\n",op->thread);
	assert(err == 0);
	op->next = Operators;
	Operators = op;
}


int open64(const char *fn, unsigned flags, unsigned mode)
{
	int ret, err;
	operator_t *op;

	init();
	ret = realopen64(fn,flags,mode);
	debug("open64(%s,%x,%x)= %d\n",fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int open(const char *fn, unsigned flags, unsigned mode)
{
	int ret, err;
	operator_t *op;

	init();
	ret = realopen(fn,flags,mode);
	debug("open(%s,%x,%x) = %d\n",fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int openat64(int at,const char *fn, unsigned flags, unsigned mode)
{
	int ret, err;
	operator_t *op;

	init();
	ret = realopenat64(at,fn,flags,mode);
	debug("openat64(%d,%s,%x,%x) = %d\n",at,fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int openat(int at,const char *fn, unsigned flags, unsigned mode)
{
	int ret, err;
	operator_t *op;

	init();
	ret = realopenat(at,fn,flags,mode);
	debug("openat(%d,%s,%x,%x) = %d\n",at,fn,flags,mode,ret);
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int close(int f)
{
	operator_t *op = Operators;
	init();
	while (op) {
		if (op->fd == f) {
			op_t *o = (op_t *) malloc(sizeof(op_t));
			o->op = cl;
			o->next = 0;
			pthread_mutex_lock(&op->lock);
			if (op->head == 0) {
				op->head = op->tail = o;
			} else {
				op->tail->next = o;
				op->tail = o;
			}
			pthread_mutex_unlock(&op->lock);
			sem_post(&op->qlen);
			debug("close(%d) -> thread %d\n",f,op->thread);
			return 0;
		}
		op = op->next;
	}
	debug("realclose(%d)\n",f);
	return realclose(f);
}


int read(int f, char *b, size_t s)
{
	init();
	debug("read(%d,%p,%d)\n",f,b,s);
	return realread(f,b,s);

}


int write(int f, const char *b, size_t s)
{
	operator_t *op = Operators;
	init();
	while (op) {
		if (op->fd == f) {
			int err;

			debug("write(%d,%p,%d) -> thread %d\n",f,b,s,op->thread);
			op_t *o = (op_t *) malloc(sizeof(op_t));
			o->op = wr;
			o->buf = (char *) malloc(s);
			memcpy(o->buf,b,s);
			o->s = s;
			o->next = 0;
			err = pthread_mutex_lock(&op->lock);
			assert(err == 0);
			if (op->head == 0) {
				op->head = op->tail = o;
			} else {
				op->tail->next = o;
				op->tail = o;
			}
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			err = sem_post(&op->qlen);
			assert(err == 0);
			return s;
		}
		op = op->next;
	}
	debug("realwrite(%d,%p,%d)\n",f,b,s);
	return realwrite(f,b,s);
}


static void *writer_thread(void *arg)
{
	operator_t *op = (operator_t *) arg;
	init();
	debug("thread %d: started\n",op->thread);
	for (;;) {
		op_t *o;
		int err;

		err = sem_wait(&op->qlen);
		assert(err == 0);
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		assert(op->head);
		o = op->head;
		if (o != op->tail) {
			op->head = o->next;
			assert(op->head);
		} else {
			assert(o->next == 0);
			op->head = op->tail = 0;
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		switch (o->op) {
		default: assert(0);
		case cl: {
				int err;
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
				err = realclose(op->fd);
				debug("thread %d: close(%d) = %d\n",op->thread,op->fd,err);
				if (err != 0)
					error("error closing file %s: %s\n",op->fn,strerror(errno));
				assert(op->head == 0);
				free(op->fn);
				op->fd = -1;
				err = pthread_mutex_unlock(&op->lock);
				assert(err == 0);
				err = sem_post(&Idle);
				assert(err == 0);
			 } break;
		case wr: {
				int err = realwrite(op->fd,o->buf,o->s);
				debug("thread %d: write(%d,%p,%d) = %d\n",op->thread,op->fd,o->buf,o->s,err);
				free(o->buf);
				if (err == -1)
					error("error writing to file %s: %s\n",op->fn,strerror(errno));
			 } break;
		case tr: {
				debug("thread %d: exit\n",op->thread);
				pthread_exit(0);
			 } break;
		}
		free(o);
	}
}


void exit(int e)
{
	init();
	debug("exit(%d)\n",e);
	operator_t *op = Operators;
	Terminate = 1;
	while (op) {
		op_t *o = (op_t *)malloc(sizeof(op_t));
		o->op = tr;
		o->next = 0;
		pthread_mutex_lock(&op->lock);
		if (op->tail)
			op->tail->next = o;
		else
			op->head = o;
		op->tail = o;
		pthread_mutex_unlock(&op->lock);
		sem_post(&op->qlen);
		op = op->next;
	}
	op = Operators;
	while (op) {
		pthread_join(op->thread,(void **)0);
		op = op->next;
	}
	debug("NumThreads = %d\n",NumThreads);
	realexit(e);
}
